#ifndef EXTRACT_ODT_TEMPLATE_H
#define EXTRACT_ODT_TEMPLATE_H

/* THIS IS AUTO-GENERATED CODE, DO NOT EDIT. */


typedef struct
{
    const char* name; /* Name of item in odt archive. */
    const char* text; /* Contents of item in odt archive. */
} odt_template_item_t;

extern const odt_template_item_t odt_template_items[];
extern int odt_template_items_num;


#endif
