/* THIS IS AUTO-GENERATED CODE, DO NOT EDIT. */

#include "odt_template.h"

const odt_template_item_t odt_template_items[] =
{
    {
        "content.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                ""
                "<office:document-content xmlns:css3t=\"http://www.w3.org/TR/css3-text/\" xmlns:grddl=\"http://www.w3.org/2003/g/data-view#\" xmlns:xhtml=\"http://www.w3.org/1999/xhtml\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:xforms=\"http://www.w3.org/2002/xforms\" xmlns:dom=\"http://www.w3.org/2001/xml-events\" xmlns:script=\"urn:oasis:names:tc:opendocument:xmlns:script:1.0\" xmlns:form=\"urn:oasis:names:tc:opendocument:xmlns:form:1.0\" xmlns:math=\"http://www.w3.org/1998/Math/MathML\" xmlns:number=\"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0\" xmlns:field=\"urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0\" xmlns:meta=\"urn:oasis:names:tc:opendocument:xmlns:meta:1.0\" xmlns:loext=\"urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0\" xmlns:officeooo=\"http://openoffice.org/2009/office\" xmlns:table=\"urn:oasis:names:tc:opendocument:xmlns:table:1.0\" xmlns:chart=\"urn:oasis:names:tc:opendocument:xmlns:chart:1.0\" xmlns:tableooo=\"http://openoffice.org/2009/table\" xmlns:draw=\"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0\" xmlns:rpt=\"http://openoffice.org/2005/report\" xmlns:dr3d=\"urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0\" xmlns:of=\"urn:oasis:names:tc:opendocument:xmlns:of:1.2\" xmlns:text=\"urn:oasis:names:tc:opendocument:xmlns:text:1.0\" xmlns:style=\"urn:oasis:names:tc:opendocument:xmlns:style:1.0\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:calcext=\"urn:org:documentfoundation:names:experimental:calc:xmlns:calcext:1.0\" xmlns:oooc=\"http://openoffice.org/2004/calc\" xmlns:drawooo=\"http://openoffice.org/2010/draw\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:ooo=\"http://openoffice.org/2004/office\" xmlns:ooow=\"http://openoffice.org/2004/writer\" xmlns:fo=\"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0\" xmlns:formx=\"urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0\" xmlns:svg=\"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" office:version=\"1.3\">"
                "<office:scripts/>"
                "<office:font-face-decls>"
                "<style:font-face style:name=\"Liberation Serif\" svg:font-family=\"&apos;Liberation Serif&apos;\" style:font-family-generic=\"roman\" style:font-pitch=\"variable\"/>"
                "<style:font-face style:name=\"Liberation Sans\" svg:font-family=\"&apos;Liberation Sans&apos;\" style:font-family-generic=\"swiss\" style:font-pitch=\"variable\"/>"
                "<style:font-face style:name=\"Unifont\" svg:font-family=\"Unifont\" style:font-family-generic=\"system\" style:font-pitch=\"variable\"/></office:font-face-decls>"
                "<office:automatic-styles/>"
                "<office:body>"
                "<office:text>"
                "<text:sequence-decls>"
                "<text:sequence-decl text:display-outline-level=\"0\" text:name=\"Illustration\"/>"
                "<text:sequence-decl text:display-outline-level=\"0\" text:name=\"Table\"/>"
                "<text:sequence-decl text:display-outline-level=\"0\" text:name=\"Text\"/>"
                "<text:sequence-decl text:display-outline-level=\"0\" text:name=\"Drawing\"/>"
                "<text:sequence-decl text:display-outline-level=\"0\" text:name=\"Figure\"/></text:sequence-decls>"
                "<text:p text:style-name=\"Standard\"/></office:text></office:body></office:document-content>"
    },

    {
        "manifest.rdf",
        "\x3c\x3f\x78\x6d\x6c\x20\x76\x65\x72\x73\x69\x6f\x6e\x3d\x22"
        "\x31\x2e\x30\x22\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x3d\x22\x75"
        "\x74\x66\x2d\x38\x22\x3f\x3e\x0a\x3c\x72\x64\x66\x3a\x52\x44\x46"
        "\x20\x78\x6d\x6c\x6e\x73\x3a\x72\x64\x66\x3d\x22\x68\x74\x74\x70"
        "\x3a\x2f\x2f\x77\x77\x77\x2e\x77\x33\x2e\x6f\x72\x67\x2f\x31\x39"
        "\x39\x39\x2f\x30\x32\x2f\x32\x32\x2d\x72\x64\x66\x2d\x73\x79\x6e"
        "\x74\x61\x78\x2d\x6e\x73\x23\x22\x3e\x0a\x20\x20\x3c\x72\x64\x66"
        "\x3a\x44\x65\x73\x63\x72\x69\x70\x74\x69\x6f\x6e\x20\x72\x64\x66"
        "\x3a\x61\x62\x6f\x75\x74\x3d\x22\x73\x74\x79\x6c\x65\x73\x2e\x78"
        "\x6d\x6c\x22\x3e\x0a\x20\x20\x20\x20\x3c\x72\x64\x66\x3a\x74\x79"
        "\x70\x65\x20\x72\x64\x66\x3a\x72\x65\x73\x6f\x75\x72\x63\x65\x3d"
        "\x22\x68\x74\x74\x70\x3a\x2f\x2f\x64\x6f\x63\x73\x2e\x6f\x61\x73"
        "\x69\x73\x2d\x6f\x70\x65\x6e\x2e\x6f\x72\x67\x2f\x6e\x73\x2f\x6f"
        "\x66\x66\x69\x63\x65\x2f\x31\x2e\x32\x2f\x6d\x65\x74\x61\x2f\x6f"
        "\x64\x66\x23\x53\x74\x79\x6c\x65\x73\x46\x69\x6c\x65\x22\x2f\x3e"
        "\x0a\x20\x20\x3c\x2f\x72\x64\x66\x3a\x44\x65\x73\x63\x72\x69\x70"
        "\x74\x69\x6f\x6e\x3e\x0a\x20\x20\x3c\x72\x64\x66\x3a\x44\x65\x73"
        "\x63\x72\x69\x70\x74\x69\x6f\x6e\x20\x72\x64\x66\x3a\x61\x62\x6f"
        "\x75\x74\x3d\x22\x22\x3e\x0a\x20\x20\x20\x20\x3c\x6e\x73\x30\x3a"
        "\x68\x61\x73\x50\x61\x72\x74\x20\x78\x6d\x6c\x6e\x73\x3a\x6e\x73"
        "\x30\x3d\x22\x68\x74\x74\x70\x3a\x2f\x2f\x64\x6f\x63\x73\x2e\x6f"
        "\x61\x73\x69\x73\x2d\x6f\x70\x65\x6e\x2e\x6f\x72\x67\x2f\x6e\x73"
        "\x2f\x6f\x66\x66\x69\x63\x65\x2f\x31\x2e\x32\x2f\x6d\x65\x74\x61"
        "\x2f\x70\x6b\x67\x23\x22\x20\x72\x64\x66\x3a\x72\x65\x73\x6f\x75"
        "\x72\x63\x65\x3d\x22\x73\x74\x79\x6c\x65\x73\x2e\x78\x6d\x6c\x22"
        "\x2f\x3e\x0a\x20\x20\x3c\x2f\x72\x64\x66\x3a\x44\x65\x73\x63\x72"
        "\x69\x70\x74\x69\x6f\x6e\x3e\x0a\x20\x20\x3c\x72\x64\x66\x3a\x44"
        "\x65\x73\x63\x72\x69\x70\x74\x69\x6f\x6e\x20\x72\x64\x66\x3a\x61"
        "\x62\x6f\x75\x74\x3d\x22\x63\x6f\x6e\x74\x65\x6e\x74\x2e\x78\x6d"
        "\x6c\x22\x3e\x0a\x20\x20\x20\x20\x3c\x72\x64\x66\x3a\x74\x79\x70"
        "\x65\x20\x72\x64\x66\x3a\x72\x65\x73\x6f\x75\x72\x63\x65\x3d\x22"
        "\x68\x74\x74\x70\x3a\x2f\x2f\x64\x6f\x63\x73\x2e\x6f\x61\x73\x69"
        "\x73\x2d\x6f\x70\x65\x6e\x2e\x6f\x72\x67\x2f\x6e\x73\x2f\x6f\x66"
        "\x66\x69\x63\x65\x2f\x31\x2e\x32\x2f\x6d\x65\x74\x61\x2f\x6f\x64"
        "\x66\x23\x43\x6f\x6e\x74\x65\x6e\x74\x46\x69\x6c\x65\x22\x2f\x3e"
        "\x0a\x20\x20\x3c\x2f\x72\x64\x66\x3a\x44\x65\x73\x63\x72\x69\x70"
        "\x74\x69\x6f\x6e\x3e\x0a\x20\x20\x3c\x72\x64\x66\x3a\x44\x65\x73"
        "\x63\x72\x69\x70\x74\x69\x6f\x6e\x20\x72\x64\x66\x3a\x61\x62\x6f"
        "\x75\x74\x3d\x22\x22\x3e\x0a\x20\x20\x20\x20\x3c\x6e\x73\x30\x3a"
        "\x68\x61\x73\x50\x61\x72\x74\x20\x78\x6d\x6c\x6e\x73\x3a\x6e\x73"
        "\x30\x3d\x22\x68\x74\x74\x70\x3a\x2f\x2f\x64\x6f\x63\x73\x2e\x6f"
        "\x61\x73\x69\x73\x2d\x6f\x70\x65\x6e\x2e\x6f\x72\x67\x2f\x6e\x73"
        "\x2f\x6f\x66\x66\x69\x63\x65\x2f\x31\x2e\x32\x2f\x6d\x65\x74\x61"
        "\x2f\x70\x6b\x67\x23\x22\x20\x72\x64\x66\x3a\x72\x65\x73\x6f\x75"
        "\x72\x63\x65\x3d\x22\x63\x6f\x6e\x74\x65\x6e\x74\x2e\x78\x6d\x6c"
        "\x22\x2f\x3e\x0a\x20\x20\x3c\x2f\x72\x64\x66\x3a\x44\x65\x73\x63"
        "\x72\x69\x70\x74\x69\x6f\x6e\x3e\x0a\x20\x20\x3c\x72\x64\x66\x3a"
        "\x44\x65\x73\x63\x72\x69\x70\x74\x69\x6f\x6e\x20\x72\x64\x66\x3a"
        "\x61\x62\x6f\x75\x74\x3d\x22\x22\x3e\x0a\x20\x20\x20\x20\x3c\x72"
        "\x64\x66\x3a\x74\x79\x70\x65\x20\x72\x64\x66\x3a\x72\x65\x73\x6f"
        "\x75\x72\x63\x65\x3d\x22\x68\x74\x74\x70\x3a\x2f\x2f\x64\x6f\x63"
        "\x73\x2e\x6f\x61\x73\x69\x73\x2d\x6f\x70\x65\x6e\x2e\x6f\x72\x67"
        "\x2f\x6e\x73\x2f\x6f\x66\x66\x69\x63\x65\x2f\x31\x2e\x32\x2f\x6d"
        "\x65\x74\x61\x2f\x70\x6b\x67\x23\x44\x6f\x63\x75\x6d\x65\x6e\x74"
        "\x22\x2f\x3e\x0a\x20\x20\x3c\x2f\x72\x64\x66\x3a\x44\x65\x73\x63"
        "\x72\x69\x70\x74\x69\x6f\x6e\x3e\x0a\x3c\x2f\x72\x64\x66\x3a\x52"
        "\x44\x46\x3e\x0a"
    },

    {
        "meta.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                ""
                "<office:document-meta xmlns:grddl=\"http://www.w3.org/2003/g/data-view#\" xmlns:meta=\"urn:oasis:names:tc:opendocument:xmlns:meta:1.0\" xmlns:ooo=\"http://openoffice.org/2004/office\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" office:version=\"1.3\">"
                "<office:meta>"
                "<meta:creation-date>2021-04-05T17:06:57.937137058</meta:creation-date>"
                "<meta:generator>LibreOffice/7.0.1.2$OpenBSD_X86_64 LibreOffice_project/00$Build-2</meta:generator>"
                "<dc:date>2021-04-06T17:14:51.409959656</dc:date>"
                "<meta:editing-duration>PT20S</meta:editing-duration>"
                "<meta:editing-cycles>1</meta:editing-cycles>"
                "<meta:document-statistic meta:table-count=\"0\" meta:image-count=\"0\" meta:object-count=\"0\" meta:page-count=\"1\" meta:paragraph-count=\"0\" meta:word-count=\"0\" meta:character-count=\"0\" meta:non-whitespace-character-count=\"0\"/></office:meta></office:document-meta>"
    },

    {
        "mimetype",
        "\x61\x70\x70\x6c\x69\x63\x61\x74\x69\x6f\x6e\x2f\x76\x6e\x64"
        "\x2e\x6f\x61\x73\x69\x73\x2e\x6f\x70\x65\x6e\x64\x6f\x63\x75\x6d"
        "\x65\x6e\x74\x2e\x74\x65\x78\x74"
    },

    {
        "settings.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                ""
                "<office:document-settings xmlns:config=\"urn:oasis:names:tc:opendocument:xmlns:config:1.0\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:ooo=\"http://openoffice.org/2004/office\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" office:version=\"1.3\">"
                "<office:settings>"
                "<config:config-item-set config:name=\"ooo:view-settings\">"
                "<config:config-item config:name=\"ViewAreaTop\" config:type=\"long\">0</config:config-item>"
                "<config:config-item config:name=\"ViewAreaLeft\" config:type=\"long\">0</config:config-item>"
                "<config:config-item config:name=\"ViewAreaWidth\" config:type=\"long\">26275</config:config-item>"
                "<config:config-item config:name=\"ViewAreaHeight\" config:type=\"long\">19502</config:config-item>"
                "<config:config-item config:name=\"ShowRedlineChanges\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"InBrowseMode\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item-map-indexed config:name=\"Views\">"
                "<config:config-item-map-entry>"
                "<config:config-item config:name=\"ViewId\" config:type=\"string\">view2</config:config-item>"
                "<config:config-item config:name=\"ViewLeft\" config:type=\"long\">4343</config:config-item>"
                "<config:config-item config:name=\"ViewTop\" config:type=\"long\">2501</config:config-item>"
                "<config:config-item config:name=\"VisibleLeft\" config:type=\"long\">0</config:config-item>"
                "<config:config-item config:name=\"VisibleTop\" config:type=\"long\">0</config:config-item>"
                "<config:config-item config:name=\"VisibleRight\" config:type=\"long\">26273</config:config-item>"
                "<config:config-item config:name=\"VisibleBottom\" config:type=\"long\">19500</config:config-item>"
                "<config:config-item config:name=\"ZoomType\" config:type=\"short\">0</config:config-item>"
                "<config:config-item config:name=\"ViewLayoutColumns\" config:type=\"short\">1</config:config-item>"
                "<config:config-item config:name=\"ViewLayoutBookMode\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ZoomFactor\" config:type=\"short\">100</config:config-item>"
                "<config:config-item config:name=\"IsSelectedFrame\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"AnchoredTextOverflowLegacy\" config:type=\"boolean\">false</config:config-item></config:config-item-map-entry></config:config-item-map-indexed></config:config-item-set>"
                "<config:config-item-set config:name=\"ooo:configuration-settings\">"
                "<config:config-item config:name=\"PrintBlackFonts\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintReversed\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"SaveThumbnail\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"EmbedFonts\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintControls\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"OutlineLevelYieldsNumbering\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintPageBackground\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"PrintAnnotationMode\" config:type=\"short\">0</config:config-item>"
                "<config:config-item config:name=\"PrintGraphics\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"EmbeddedDatabaseName\" config:type=\"string\"/>"
                "<config:config-item config:name=\"ProtectForm\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintLeftPages\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"PrintProspect\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintHiddenText\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintRightPages\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"PrintFaxName\" config:type=\"string\"/>"
                "<config:config-item config:name=\"PrintPaperFromSetup\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"TabsRelativeToIndent\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"RedlineProtectionKey\" config:type=\"base64Binary\"/>"
                "<config:config-item config:name=\"PrintTextPlaceholder\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"MathBaselineAlignment\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"ProtectBookmarks\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"IgnoreTabsAndBlanksForLineCalculation\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ContinuousEndnotes\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"FieldAutoUpdate\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"EmptyDbFieldHidesPara\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"ApplyParagraphMarkFormatToNumbering\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintEmptyPages\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"TabOverMargin\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"EmbedAsianScriptFonts\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"EmbedLatinScriptFonts\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"DisableOffPagePositioning\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"EmbedOnlyUsedFonts\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"MsWordCompMinLineHeightByFly\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"SurroundTextWrapSmall\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"BackgroundParaOverDrawings\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ClippedPictures\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"FloattableNomargins\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"UnbreakableNumberings\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"EmbedSystemFonts\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"TabOverflow\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"PrintTables\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"PrintDrawings\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"ConsiderTextWrapOnObjPos\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintSingleJobs\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"SmallCapsPercentage66\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"CollapseEmptyCellPara\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"HeaderSpacingBelowLastPara\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"RsidRoot\" config:type=\"int\">2052946</config:config-item>"
                "<config:config-item config:name=\"PrinterSetup\" config:type=\"base64Binary\"/>"
                "<config:config-item config:name=\"CurrentDatabaseCommand\" config:type=\"string\"/>"
                "<config:config-item config:name=\"AlignTabStopPosition\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"ClipAsCharacterAnchoredWriterFlyFrames\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"DoNotCaptureDrawObjsOnPage\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"SaveGlobalDocumentLinks\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"CurrentDatabaseCommandType\" config:type=\"int\">0</config:config-item>"
                "<config:config-item config:name=\"LoadReadonly\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"DoNotResetParaAttrsForNumFont\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"StylesNoDefault\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"LinkUpdateMode\" config:type=\"short\">1</config:config-item>"
                "<config:config-item config:name=\"DoNotJustifyLinesWithManualBreak\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PropLineSpacingShrinksFirstLine\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"TabAtLeftIndentForParagraphsInList\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ProtectFields\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"UnxForceZeroExtLeading\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"CurrentDatabaseDataSource\" config:type=\"string\"/>"
                "<config:config-item config:name=\"UseFormerTextWrapping\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"UseFormerLineSpacing\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"AllowPrintJobCancel\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"SubtractFlysAnchoredAtFlys\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"AddParaSpacingToTableCells\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"AddExternalLeading\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"Rsid\" config:type=\"int\">2178852</config:config-item>"
                "<config:config-item config:name=\"AddVerticalFrameOffsets\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"TreatSingleColumnBreakAsPageBreak\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"AddFrameOffsets\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"IsLabelDocument\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"MsWordCompTrailingBlanks\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrinterPaperFromSetup\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"IgnoreFirstLineIndentInNumbering\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrinterName\" config:type=\"string\"/>"
                "<config:config-item config:name=\"IsKernAsianPunctuation\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrinterIndependentLayout\" config:type=\"string\">high-resolution</config:config-item>"
                "<config:config-item config:name=\"TableRowKeep\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"UpdateFromTemplate\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"EmbedComplexScriptFonts\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"UseOldPrinterMetrics\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"InvertBorderSpacing\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"PrintProspectRTL\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ApplyUserData\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"AddParaTableSpacingAtStart\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"SaveVersionOnClose\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"CharacterCompressionType\" config:type=\"short\">0</config:config-item>"
                "<config:config-item config:name=\"UseOldNumbering\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"UseFormerObjectPositioning\" config:type=\"boolean\">false</config:config-item>"
                "<config:config-item config:name=\"ChartAutoUpdate\" config:type=\"boolean\">true</config:config-item>"
                "<config:config-item config:name=\"AddParaTableSpacing\" config:type=\"boolean\">true</config:config-item></config:config-item-set></office:settings></office:document-settings>"
    },

    {
        "styles.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                ""
                "<office:document-styles xmlns:css3t=\"http://www.w3.org/TR/css3-text/\" xmlns:grddl=\"http://www.w3.org/2003/g/data-view#\" xmlns:xhtml=\"http://www.w3.org/1999/xhtml\" xmlns:dom=\"http://www.w3.org/2001/xml-events\" xmlns:script=\"urn:oasis:names:tc:opendocument:xmlns:script:1.0\" xmlns:form=\"urn:oasis:names:tc:opendocument:xmlns:form:1.0\" xmlns:math=\"http://www.w3.org/1998/Math/MathML\" xmlns:number=\"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0\" xmlns:field=\"urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0\" xmlns:meta=\"urn:oasis:names:tc:opendocument:xmlns:meta:1.0\" xmlns:loext=\"urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0\" xmlns:officeooo=\"http://openoffice.org/2009/office\" xmlns:table=\"urn:oasis:names:tc:opendocument:xmlns:table:1.0\" xmlns:chart=\"urn:oasis:names:tc:opendocument:xmlns:chart:1.0\" xmlns:tableooo=\"http://openoffice.org/2009/table\" xmlns:draw=\"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0\" xmlns:rpt=\"http://openoffice.org/2005/report\" xmlns:dr3d=\"urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0\" xmlns:of=\"urn:oasis:names:tc:opendocument:xmlns:of:1.2\" xmlns:text=\"urn:oasis:names:tc:opendocument:xmlns:text:1.0\" xmlns:style=\"urn:oasis:names:tc:opendocument:xmlns:style:1.0\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:calcext=\"urn:org:documentfoundation:names:experimental:calc:xmlns:calcext:1.0\" xmlns:oooc=\"http://openoffice.org/2004/calc\" xmlns:drawooo=\"http://openoffice.org/2010/draw\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:ooo=\"http://openoffice.org/2004/office\" xmlns:ooow=\"http://openoffice.org/2004/writer\" xmlns:fo=\"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0\" xmlns:svg=\"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" office:version=\"1.3\">"
                "<office:font-face-decls>"
                "<style:font-face style:name=\"Liberation Serif\" svg:font-family=\"&apos;Liberation Serif&apos;\" style:font-family-generic=\"roman\" style:font-pitch=\"variable\"/>"
                "<style:font-face style:name=\"Liberation Sans\" svg:font-family=\"&apos;Liberation Sans&apos;\" style:font-family-generic=\"swiss\" style:font-pitch=\"variable\"/>"
                "<style:font-face style:name=\"Unifont\" svg:font-family=\"Unifont\" style:font-family-generic=\"system\" style:font-pitch=\"variable\"/></office:font-face-decls>"
                "<office:styles>"
                "<style:default-style style:family=\"graphic\">"
                "<style:graphic-properties svg:stroke-color=\"#3465a4\" draw:fill-color=\"#729fcf\" fo:wrap-option=\"no-wrap\" draw:shadow-offset-x=\"0.1181in\" draw:shadow-offset-y=\"0.1181in\" draw:start-line-spacing-horizontal=\"0.1114in\" draw:start-line-spacing-vertical=\"0.1114in\" draw:end-line-spacing-horizontal=\"0.1114in\" draw:end-line-spacing-vertical=\"0.1114in\" style:flow-with-text=\"false\"/>"
                "<style:paragraph-properties style:text-autospace=\"ideograph-alpha\" style:line-break=\"strict\" style:font-independent-line-spacing=\"false\">"
                "<style:tab-stops/></style:paragraph-properties>"
                "<style:text-properties style:use-window-font-color=\"true\" loext:opacity=\"0%\" style:font-name=\"Liberation Serif\" fo:font-size=\"12pt\" fo:language=\"en\" fo:country=\"US\" style:letter-kerning=\"true\" style:font-name-asian=\"Unifont\" style:font-size-asian=\"10.5pt\" style:language-asian=\"zh\" style:country-asian=\"CN\" style:font-name-complex=\"Unifont\" style:font-size-complex=\"12pt\" style:language-complex=\"hi\" style:country-complex=\"IN\"/></style:default-style>"
                "<style:default-style style:family=\"paragraph\">"
                "<style:paragraph-properties fo:orphans=\"2\" fo:widows=\"2\" fo:hyphenation-ladder-count=\"no-limit\" style:text-autospace=\"ideograph-alpha\" style:punctuation-wrap=\"hanging\" style:line-break=\"strict\" style:tab-stop-distance=\"0.4925in\" style:writing-mode=\"page\"/>"
                "<style:text-properties style:use-window-font-color=\"true\" loext:opacity=\"0%\" style:font-name=\"Liberation Serif\" fo:font-size=\"12pt\" fo:language=\"en\" fo:country=\"US\" style:letter-kerning=\"true\" style:font-name-asian=\"Unifont\" style:font-size-asian=\"10.5pt\" style:language-asian=\"zh\" style:country-asian=\"CN\" style:font-name-complex=\"Unifont\" style:font-size-complex=\"12pt\" style:language-complex=\"hi\" style:country-complex=\"IN\" fo:hyphenate=\"false\" fo:hyphenation-remain-char-count=\"2\" fo:hyphenation-push-char-count=\"2\" loext:hyphenation-no-caps=\"false\"/></style:default-style>"
                "<style:default-style style:family=\"table\">"
                "<style:table-properties table:border-model=\"collapsing\"/></style:default-style>"
                "<style:default-style style:family=\"table-row\">"
                "<style:table-row-properties fo:keep-together=\"auto\"/></style:default-style>"
                "<style:style style:name=\"Standard\" style:family=\"paragraph\" style:class=\"text\">"
                "<style:paragraph-properties fo:margin-top=\"0in\" fo:margin-bottom=\"0.0799in\" style:contextual-spacing=\"false\"/></style:style>"
                "<style:style style:name=\"Heading\" style:family=\"paragraph\" style:parent-style-name=\"Standard\" style:next-style-name=\"Text_20_body\" style:class=\"text\">"
                "<style:paragraph-properties fo:margin-top=\"0.1665in\" fo:margin-bottom=\"0.0835in\" style:contextual-spacing=\"false\" fo:keep-with-next=\"always\"/>"
                "<style:text-properties style:font-name=\"Liberation Sans\" fo:font-family=\"&apos;Liberation Sans&apos;\" style:font-family-generic=\"swiss\" style:font-pitch=\"variable\" fo:font-size=\"14pt\" style:font-name-asian=\"Unifont\" style:font-family-asian=\"Unifont\" style:font-family-generic-asian=\"system\" style:font-pitch-asian=\"variable\" style:font-size-asian=\"14pt\" style:font-name-complex=\"Unifont\" style:font-family-complex=\"Unifont\" style:font-family-generic-complex=\"system\" style:font-pitch-complex=\"variable\" style:font-size-complex=\"14pt\"/></style:style>"
                "<style:style style:name=\"Text_20_body\" style:display-name=\"Text body\" style:family=\"paragraph\" style:parent-style-name=\"Standard\" style:class=\"text\">"
                "<style:paragraph-properties fo:margin-top=\"0in\" fo:margin-bottom=\"0.0972in\" style:contextual-spacing=\"false\" fo:line-height=\"115%\"/></style:style>"
                "<style:style style:name=\"List\" style:family=\"paragraph\" style:parent-style-name=\"Text_20_body\" style:class=\"list\">"
                "<style:text-properties style:font-size-asian=\"12pt\"/></style:style>"
                "<style:style style:name=\"Caption\" style:family=\"paragraph\" style:parent-style-name=\"Standard\" style:class=\"extra\">"
                "<style:paragraph-properties fo:margin-top=\"0.0835in\" fo:margin-bottom=\"0.0835in\" style:contextual-spacing=\"false\" text:number-lines=\"false\" text:line-number=\"0\"/>"
                "<style:text-properties fo:font-size=\"12pt\" fo:font-style=\"italic\" style:font-size-asian=\"12pt\" style:font-style-asian=\"italic\" style:font-size-complex=\"12pt\" style:font-style-complex=\"italic\"/></style:style>"
                "<style:style style:name=\"Index\" style:family=\"paragraph\" style:parent-style-name=\"Standard\" style:class=\"index\">"
                "<style:paragraph-properties text:number-lines=\"false\" text:line-number=\"0\"/>"
                "<style:text-properties style:font-size-asian=\"12pt\"/></style:style>"
                "<text:outline-style style:name=\"Outline\">"
                "<text:outline-level-style text:level=\"1\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"2\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"3\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"4\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"5\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"6\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"7\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"8\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"9\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style>"
                "<text:outline-level-style text:level=\"10\" style:num-format=\"\">"
                "<style:list-level-properties text:list-level-position-and-space-mode=\"label-alignment\">"
                "<style:list-level-label-alignment text:label-followed-by=\"listtab\"/></style:list-level-properties></text:outline-level-style></text:outline-style>"
                "<text:notes-configuration text:note-class=\"footnote\" style:num-format=\"1\" text:start-value=\"0\" text:footnotes-position=\"page\" text:start-numbering-at=\"document\"/>"
                "<text:notes-configuration text:note-class=\"endnote\" style:num-format=\"i\" text:start-value=\"0\"/>"
                "<text:linenumbering-configuration text:number-lines=\"false\" text:offset=\"0.1965in\" style:num-format=\"1\" text:number-position=\"left\" text:increment=\"5\"/></office:styles>"
                "<office:automatic-styles>"
                "<style:page-layout style:name=\"Mpm1\">"
                "<style:page-layout-properties fo:page-width=\"8.5in\" fo:page-height=\"11in\" style:num-format=\"1\" style:print-orientation=\"portrait\" fo:margin-top=\"0.7874in\" fo:margin-bottom=\"0.7874in\" fo:margin-left=\"0.7874in\" fo:margin-right=\"0.7874in\" style:writing-mode=\"lr-tb\" style:layout-grid-color=\"#c0c0c0\" style:layout-grid-lines=\"20\" style:layout-grid-base-height=\"0.278in\" style:layout-grid-ruby-height=\"0.139in\" style:layout-grid-mode=\"none\" style:layout-grid-ruby-below=\"false\" style:layout-grid-print=\"false\" style:layout-grid-display=\"false\" style:footnote-max-height=\"0in\">"
                "<style:footnote-sep style:width=\"0.0071in\" style:distance-before-sep=\"0.0398in\" style:distance-after-sep=\"0.0398in\" style:line-style=\"solid\" style:adjustment=\"left\" style:rel-width=\"25%\" style:color=\"#000000\"/></style:page-layout-properties>"
                "<style:header-style/>"
                "<style:footer-style/></style:page-layout></office:automatic-styles>"
                "<office:master-styles>"
                "<style:master-page style:name=\"Standard\" style:page-layout-name=\"Mpm1\"/></office:master-styles></office:document-styles>"
    },

    {
        "META-INF/manifest.xml",
        ""
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                ""
                "<manifest:manifest xmlns:manifest=\"urn:oasis:names:tc:opendocument:xmlns:manifest:1.0\" manifest:version=\"1.3\" xmlns:loext=\"urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0\">\n"
                " "
                "<manifest:file-entry manifest:full-path=\"/\" manifest:version=\"1.3\" manifest:media-type=\"application/vnd.oasis.opendocument.text\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"meta.xml\" manifest:media-type=\"text/xml\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"settings.xml\" manifest:media-type=\"text/xml\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"Configurations2/\" manifest:media-type=\"application/vnd.sun.xml.ui.configuration\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"manifest.rdf\" manifest:media-type=\"application/rdf+xml\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"styles.xml\" manifest:media-type=\"text/xml\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"content.xml\" manifest:media-type=\"text/xml\"/>\n"
                " "
                "<manifest:file-entry manifest:full-path=\"Thumbnails/thumbnail.png\" manifest:media-type=\"image/png\"/>\n"
                "</manifest:manifest>"
    },

    {
        "Thumbnails/thumbnail.png",
        "\x89\x50\x4e\x47\x0d\x0a\x1a\x0a\x00\x00\x00\x0d\x49\x48\x44"
        "\x52\x00\x00\x00\xc6\x00\x00\x01\x00\x08\x03\x00\x00\x00\xdf\x83"
        "\xf9\x72\x00\x00\x00\x09\x50\x4c\x54\x45\xff\xff\xff\x00\x00\x00"
        "\xff\xff\xff\x7e\xef\x8f\x4f\x00\x00\x00\x48\x49\x44\x41\x54\x78"
        "\xda\xed\xc1\x31\x01\x00\x00\x00\xc2\xa0\xf5\x4f\x6d\x0c\x1f\xa0"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x78\x18\xc7\x00\x00\x01\xf9\xd2\xb5\x9a\x00\x00\x00\x00\x49"
        "\x45\x4e\x44\xae\x42\x60\x82"
    },

};

int odt_template_items_num = 8;
