#ifndef EXTRACT_DOCX_TEMPLATE_H
#define EXTRACT_DOCX_TEMPLATE_H

/* THIS IS AUTO-GENERATED CODE, DO NOT EDIT. */


typedef struct
{
    const char* name; /* Name of item in docx archive. */
    const char* text; /* Contents of item in docx archive. */
} docx_template_item_t;

extern const docx_template_item_t docx_template_items[];
extern int docx_template_items_num;


#endif
